/**
 * isolate -- Isolate a user process by limiting its access to the operating
 * system and the environment.
 *
 * By Chris Palmer <chris@noncombatant.org>
 *
 * $Id: isolate.cpp 22 2010-01-07 00:56:02Z snackypants $
 *
 * Released under the terms of the GNU General Public License, version 2.
 */


#define _ATFILE_SOURCE

#include <sys/param.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <gelf.h>
#ifdef linux
      #include <grp.h>    // For setgroups(2)
#endif
#include <libgen.h>
#include <regex.h>
#include <signal.h>
#define _WITH_GETLINE
#include <stdio.h>
#include <unistd.h>

#include <cstring>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <stdexcept>

using namespace std;

#include "configuration.h"
#include "help.h"
#include "privilege.h"


extern char ** environ;

typedef set<string> string_set;


/* Global variables. */

static string confinement_path;
static uid_t isolator, invoker;
static bool x11_authentication_needed = false;


/* Functions. */

DIR * fdopendir(int fd);

int unlink_recursively(const char * pathname, int parent) {

      DIR * dir = NULL;
      int fd = -1;

      /*
       * First try to unlink it. If we fail because of EISDIR
       * (Linux) or EPERM (which FreeBSD overloads to mean
       * EISDIR in this case), then assume it's a directory and
       * recurse into it.
       */
      #if defined __FreeBSD__
            #define eisdir EPERM
      #elif defined __linux__
            #define eisdir EISDIR
      #else
            #error "Unsupported platform"
      #endif
      if (0 == unlinkat(parent, pathname, 0))
            return 0;

      if (eisdir != errno)
            goto error;


      /*
       * Check our assumption about this being a directory. Crucially, do
       * not follow symlinks.
       */
      fd = openat(parent, pathname, O_RDONLY | O_NOFOLLOW);
      if (-1 == fd)
            goto error;

      struct stat st;
      if (fstat(fd, &st))
            goto error;

      if (! S_ISDIR(st.st_mode))
            goto error;


      /*
       * Now unlink everything in the directory, recursing as needed.
       */
      dir = fdopendir(fd);
      if (! dir)
            goto error;

      struct dirent * ent;
      while ( (ent = readdir(dir)) ) {
            if (0 == strcmp(".", ent->d_name) || 0 == strcmp("..", ent->d_name))
                  continue;

            if (unlink_recursively(ent->d_name, fd))
                  goto error;
      }

      if (unlinkat(parent, pathname, AT_REMOVEDIR))
            goto error;

      if (closedir(dir))
            goto error;

      return 0;

error:
      cerr << pathname << ": " << strerror(errno) << endl;
      if (dir)
            (void) closedir(dir);
      else
            (void) close(fd);
      return -1;
}


/**
 * NOTE: The implementation of this function is amusing.
 *
 * @param n The number to convert to a string.
 *
 * @param base The base in which to represent the number. Can be 8, 10, or
 * 16. Any other value is treated as if it were 10.
 *
 * @return A pointer to a static array of characters representing the number
 * as a string.
 */
static const char * to_string(long long int n, unsigned char base) {

#define bfr_sz 32

      /* We are sure to not overflow this, given our use of snprintf below. */
      /* Flawfinder: ignore */
      static char bfr[bfr_sz];

      const char * frmt = "%lld";
      if (16 == base) {
            frmt = "%llx";
      } else if (8 == base) {
            frmt = "%llo";
      }

      /* Flawfinder: ignore */
      snprintf(bfr, bfr_sz, frmt, n);
      bfr[bfr_sz - 1] = '\0';
      return bfr;
}


/**
 * @param path The path to check.
 *
 * @returns true if the path is absolute. Note that our definition of
 * "absolute" includes paths beginning with "./", since they are "absolute"
 * relative to a known path. Paths beginning with "../" are not absolute,
 * nor are paths that do not begin with "/" or "./" such as "ls" or
 * "foo.txt".
 */
static bool is_absolute_path(const string & path) throw () {
      return '/' == path[0]
             || ('.' == path[0] && '/' == path[1]);
}


/**

 * @param file The basename of a file to find in one of the directories
 * specified in $PATH. (If $PATH is not available, DEFAULT_PATH is used.)
 *
 * @return A new string giving the full path in which file was found.
 *
 * @throws range_error if file was not found in $PATH/DEFAULT_PATH.
 */
static string which(const string & file) throw (runtime_error) {
      assert(geteuid() != 0 && getegid() != 0);

      if (is_absolute_path(file)) {
            return file;
      }

      string pth = DEFAULT_PATH;
      char *p = getenv("PATH");
      if (p) {
            pth = p;
      }

      string::size_type i = 0;
      string::size_type f = pth.find(':', i);
      do {
            string s = pth.substr(i, f - i) + "/" + file;

            /* We are using access(2) as a simple heuristic to see if we
             * should bother return s or keep searching. We are not using
             * it as a critical security check. */
            /* Flawfinder: ignore */
            if (0 == access(s.c_str(), X_OK)) {
                  return s;
            }

            i = f + 1;
            f = pth.find(':', i);
      } while (string::npos != f);

      if (! is_absolute_path(file)) {
            throw runtime_error("File not found in $PATH");
      }

      /* Flawfinder: ignore */
      if (0 == access(file.c_str(), X_OK)) {
            return file;
      }

      throw runtime_error("File not found in $PATH");
}


/**
 * @param matchee The string to match against pattern.
 * @param pattern The pattern to test.
 * @param flags The regular expression compilation flags (see regex(3)).
 * (REG_EXTENDED|REG_NOSUB is applied implicitly.)
 *
 * @return true if matchee matches pattern, false otherwise.
 */
static bool matches_pattern(const string & matchee, const char * pattern, int flags)
      throw (logic_error)
{
      regex_t xprsn;
      if (regcomp(&xprsn, pattern, flags|REG_EXTENDED|REG_NOSUB)) {
            throw logic_error("Failed to compile regular expression");
      }

      if (0 == regexec(&xprsn, matchee.c_str(), 1, NULL, 0)) {
            regfree(&xprsn);
            return true;
      }

      regfree(&xprsn);
      return false;
}


/**
 * NOTE: The regular expression allows e.g. "......:0", which is illegal.
 * Our purpose here is to make sure that display is safe to interpolate in a
 * command we send to system(3), not to make sure it is 100% true and good.
 * Safe but illegal/incorrect/wacky display names will cause xauth to fail
 * (e.g. "host not found"). I consider that okay: isolate will give users
 * with bad $DISPLAYs the same results they would have gotten without
 * isolate.
 *
 * @param display The X11 display name to check.
 *
 * @return true if the display name is valid, false otherwise.
static bool is_safe_x11_display(const string & display) throw (logic_error) {
      return matches_pattern(display, "^[.0-9a-z_-]*:[0-9]+\\.?[0-9]*$", REG_ICASE);
}
 */


/**
 * @throws runtime_error If CONFINEMENT_ROOT is not owned by UID and GID
 * 0, if it exists but does not have CONFINEMENT_ROOT_MODE, or if
 * CONFINEMENT_ROOT cannot be created.
 */
static void create_confinement_root() throw (runtime_error) {
      struct stat stt;

      if (0 == stat(CONFINEMENT_ROOT.c_str(), &stt)) {
            if (0 != stt.st_uid || 0 != stt.st_gid) {
                  throw runtime_error(CONFINEMENT_ROOT + " not owned by 0:0. Cannot continue.");
            }

            if (stt.st_mode != CONFINEMENT_ROOT_MODE) {
                  string is = to_string(stt.st_mode, 8);
                  string shld = to_string(CONFINEMENT_ROOT_MODE, 8);
                  throw runtime_error(CONFINEMENT_ROOT + " is mode " + is + "; should be " + shld);
            }
      }
      else if (ENOENT == errno) {
            /* Make sure we have the right EGID so that the directory has
             * the right group ownership. */
            gid_t egid = getegid();
            setegid(0);

            if (mkdir(CONFINEMENT_ROOT.c_str(), CONFINEMENT_ROOT_MODE)) {
                  throw runtime_error("Could not create " + CONFINEMENT_ROOT);
            }

            setegid(egid);
      }
      else {
            throw runtime_error(CONFINEMENT_ROOT + ": " + strerror(errno));
      }
}


static map<int, string> resources;
static void populate_resources() {
      resources[RLIMIT_AS]      = "RLIMIT_AS (virtual memory size)";
      resources[RLIMIT_CORE]    = "RLIMIT_CORE (core file size)";
      resources[RLIMIT_DATA]    = "RLIMIT_DATA (data segment size)";
      resources[RLIMIT_NOFILE]  = "RLIMIT_NOFILE (number of files)";
      resources[RLIMIT_MEMLOCK] = "RLIMIT_MEMLOCK (locked memory size)";
#ifndef linux
      resources[RLIMIT_SBSIZE]  = "RLIMIT_SBSIZE (socket buffer size)";
#endif
      resources[RLIMIT_NPROC]   = "RLIMIT_NPROC (simultaneous processes)";
      resources[RLIMIT_RSS]     = "RLIMIT_RSS (resident set size)";
      resources[RLIMIT_STACK]   = "RLIMIT_STACK (stack size)";
      resources[RLIMIT_CPU]     = "RLIMIT_CPU (CPU time)";
      resources[RLIMIT_FSIZE]   = "RLIMIT_FSIZE (file size)";
}

/**
 * @param resource The resource to set the limit on. (See setrlimit(2).)
 *
 * @param limit The value -- hard and soft -- of the limit.
 *
 * @throws runtime_error If setrlimit fails.
 */
static void set_resource_limit(const int resource, const rlim_t limit) throw (runtime_error) {
      assert(geteuid() == invoker && getegid() == invoker);

      struct rlimit rlmt = { limit, limit };
      if (setrlimit(resource, &rlmt)) {
            populate_resources();
            throw runtime_error("Could not set resource " + resources[resource]
                  + " to " + to_string(limit, 10) + ": " + strerror(errno));
      }
}


/**
 * Sends signal 0 to all processes running as isolator. This function
 * returns only if that failed because of ESRCH -- i.e. if no processes were
 * running as isolator.
 *
 * NOTE: This technique appears not to work on Linux -- kill(-1, 0) always
 * returns 0.
 *
 * NOTE: This function is of only limited usefulness, since there is an
 * inherent race condition.  There is no real way to ensure that nobody is
 * running with the same UID, or to lock other UIDs from spawning processes.
 *
 * @throws runtime_error If it was impossible to fork, or if other
 * processes had been running as isolator.
 */
static void ensure_no_processes() throw (runtime_error) {
#ifndef linux
      int sts = 0;
      errno = 0;
      pid_t pid = fork();

      if (-1 == pid) {
            throw runtime_error("Could not fork!");
      }

      if (0 == pid) {
            drop_privilege_permanently(isolator);

            if (0 == kill(-1, 0)) {
                  _exit(1);
            } else if (ESRCH != errno) {
                  _exit(errno);
            }

            _exit(0);
      }
      else {
            (void) wait(&sts);
      }

      if (0 != WEXITSTATUS(sts)) {
            throw runtime_error(string("Could not ensure that no processes were running as UID ")
                  + to_string(isolator, 10));
      }
#endif
}


/**
 * @return An unpredictable (uid_t is not large enough for me to say
 * "secure") UID.
 *
 * @throws runtime_error If a security assertion cannot be met.
 */
static uid_t random_uid() throw (runtime_error) {
      uid_t u;
      for (unsigned char i = 0; i < 10; i++) {
#ifdef linux
            
            int rndm = open(DEV_RANDOM.c_str(), O_RDONLY);
            ssize_t r = read(rndm, reinterpret_cast<char *>(&u), sizeof(u));
            close(rndm);
            if (sizeof(u) != r) {
                  continue;
            }
#else
            u = arc4random();
#endif

            if (u > 0xFFFF) {
                  return u;
            }
      }

      throw runtime_error("Could not generate a good random UID after 10 attempts. Bummer!");
}


/**
 * @param file The file to test.
 *
 * @return True if the file's first two bytes are "#!". Advances the file
 * pointer by two bytes.
 */
static bool shell_magic(FILE *file) throw () {
      /* Flawfinder: ignore */
      char bfr [2];
      if (2 != fread(bfr, sizeof(char), 2, file)) {
            return false;
      }

      if ('#' == bfr[0] && '!' == bfr[1]) {
            return true;
      }

      return false;
}


/**
 * @param name A string that might be the name of a code library, e.g.
 * libc.so.7.
 *
 * @return Whether or not the string is plausibly the name of a code
 * library.
 */
static bool names_library(const string & name) throw() {
    return matches_pattern(name, "^lib.+\\.so[.0-9]*$", 0);
}


/**
 * @param elf A pointer to an Elf object (see elf(3)).
 *
 * @return A pair of vectors of strings. The first member is a vector of
 * libraries (e.g. "libc.so.7"), and the second is a vector of search paths
 * (e.g. "/usr/local/gadget/lib").
 *
 * @throws runtime_error If the Elf could not be parsed normally.
 */
static pair<string_set *, string_set *> * dynamic_loads(Elf *elf) throw (runtime_error) {
      assert(geteuid() != 0 && getegid() != 0);

      GElf_Ehdr ehdr;
      if (! gelf_getehdr(elf, &ehdr)) {
            throw runtime_error(string("elf_getehdr failed: ") + elf_errmsg(-1));
      }

      Elf_Scn *scn = elf_nextscn(elf, NULL);
      GElf_Shdr shdr;
      string to_find = ".dynstr";
      while (scn) {
            if (NULL == gelf_getshdr(scn, &shdr)) {
                  throw runtime_error(string("getshdr() failed: ") + elf_errmsg(-1));
            }

            char * nm = elf_strptr(elf, ehdr.e_shstrndx, shdr.sh_name);
            if (NULL == nm) {
                  throw runtime_error(string("elf_strptr() failed: ") + elf_errmsg(-1));
            }

            if (to_find == nm) {
                  break;
            }

            scn = elf_nextscn(elf, scn);
      }

      Elf_Data *data = NULL;
      size_t n = 0;
      string_set *lbrrs = new string_set(), *pths = new string_set();

      pths->insert("/lib");
      pths->insert("/usr/lib");
      pths->insert("/usr/local/lib");

      while (n < shdr.sh_size && (data = elf_getdata(scn, data)) ) {
            char *bfr = static_cast<char *>(data->d_buf);
            char *p = bfr + 1;
            while (p < bfr + data->d_size) {
                  if (names_library(p)) {
                        lbrrs->insert(p);
                        if (string(p).substr(0, 9) == "libX11.so") {
                              x11_authentication_needed = true;
                        }
                  } else if ('/' == *p) {
                        pths->insert(p);
                  }

                  size_t lngth = strlen(p) + 1;
                  n += lngth;
                  p += lngth;
            }
      }

      return new pair<string_set *, string_set *>(lbrrs, pths);
}


static string_set already_copied;

/**
 * @param image The path to the executable image file.
 *
 * @return The number of dependencies copied into confinement_path.  XXX:
 * This allows a malicious isolatee to increase the number of files it can
 * open. However, I think it is all under the control of the loader?
 *
 * @throws runtime_error if the image could not be parsed. Does not throw
 * if a dependency could not be copied.
 */
static rlim_t count_dependencies(const string & image) throw (runtime_error) {
      if (EV_NONE == elf_version(EV_CURRENT)) {
            throw runtime_error(string("ELF library initialization failed: ") + elf_errmsg(-1));
      }

      int fl = open(image.c_str(), O_RDONLY);
      if (-1 == fl) {
            throw runtime_error(string("Could not open ") + image + ": " + strerror(errno));
      }

      Elf *elf = elf_begin(fl, ELF_C_READ, NULL);
      if (NULL == elf) {
            throw runtime_error(string("elf_begin failed: ") + elf_errmsg(-1));
      }

      if (ELF_K_ELF != elf_kind(elf)) {
            throw runtime_error(image + " is not an ELF object.");
      }

      pair<string_set *, string_set *> *r = dynamic_loads(elf);
      string_set lds = *r->first;
      rlim_t lmt_fls = 0;

      for (string_set::iterator ld = lds.begin(); ld != lds.end(); ++ld) {
            string_set pths = *r->second;
            bool fnd = false;

            for (string_set::iterator pth = pths.begin(); pth != pths.end(); ++pth) {
                  string src = *pth + '/' + *ld;
                  if (already_copied.count(src)) {
                        fnd = true;
                        continue;
                  }
                  string dstntn = confinement_path + src;

                  try {
                        lmt_fls += 1;
                        fnd = true;
                        already_copied.insert(src);
                        lmt_fls += count_dependencies(src);    // FIXME: Grossly inefficient.
                        break;
                  } catch (runtime_error e) { }
            }

            if (! fnd) {
                  cerr << "Could not find library " << *ld << endl;
            }
      }

      elf_end(elf);
      close(fl);
      return lmt_fls;
}


/**
 * Prints the resource usage of the child process to cerr.
 */
static void print_rusage() throw () {
      struct rusage usg;
      if (getrusage(RUSAGE_CHILDREN, &usg)) {
            cerr << "Could not get child's resource usage: " << strerror(errno) << endl;
            return;
      }

      cerr << "User time: " << usg.ru_utime.tv_sec << "." << setfill('0') << setw(6)
           << usg.ru_utime.tv_usec << endl
           << "System time: " << usg.ru_stime.tv_sec << "." << setfill('0') << setw(6)
           << usg.ru_stime.tv_usec << endl
           << "Unshared data size: " << usg.ru_idrss << endl
           << "Unshared stack size: " << usg.ru_isrss << endl;
}


/**
 * @param resource The resource for which to find the limit. See
 * getrlimit(2).
 *
 * @return The current process' maximum limit for the given resource.
 */
static rlim_t get_resource_limit(int resource) throw () {
      struct rlimit rlmt;
      if (getrlimit(resource, &rlmt)) {
            throw runtime_error(string("getrlimit: ") + strerror(errno));
      }
      return rlmt.rlim_max;
}


int main(int argc, char * argv []) {

      /*
       * Discover the X authorization.
       */
      string dsply = "", xauthrty = "";
      //if (x11_authentication_needed) {
            char * p = getenv("DISPLAY");
            if (p)
                  dsply = string(p);
            p = getenv("XAUTHORITY");
            if (p)
                  xauthrty = string(p);
      //}

      environ = NULL;


      /*
       * Set default limits.
       */

      rlim_t
            lmt_all = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_AS)),
            lmt_cr = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_CORE)),
            lmt_dta = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_DATA)),
            lmt_fls = min(static_cast<rlim_t>(3), get_resource_limit(RLIMIT_NOFILE)),
            lmt_fl_sz = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_FSIZE)),
            lmt_mlck = min(static_cast<rlim_t>(0x200000), get_resource_limit(RLIMIT_MEMLOCK)),
#ifndef linux
            lmt_ntwrk = min(static_cast<rlim_t>(0x200000), get_resource_limit(RLIMIT_SBSIZE)),
#endif
            lmt_prcs = min(static_cast<rlim_t>(0), get_resource_limit(RLIMIT_NPROC)),
            lmt_rss = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_RSS)),
            lmt_stck = min(DEFAULT_MEMORY_LIMIT, get_resource_limit(RLIMIT_STACK)),
            lmt_tm = min(static_cast<rlim_t>(10), get_resource_limit(RLIMIT_CPU))
            ;

      bool vrbs = false;


      /*
       * Parse command line.
       */

      char c;
      while (-1 != (c = getopt(argc, argv, "a:c:d:f:hm:n:p:r:s:t:vz:"))) {
            switch (c) {
            case 'a':
                  lmt_all = strtoll(optarg, NULL, 0);
                  break;
            case 'c':
                  lmt_cr = strtoll(optarg, NULL, 0);
                  break;
            case 'd':
                  lmt_dta = strtoll(optarg, NULL, 0);
                  break;
            case 'f':
                  lmt_fls = strtoll(optarg, NULL, 0);
                  break;
            case 'm':
                  lmt_mlck = strtoll(optarg, NULL, 0);
                  break;
#ifndef linux
            case 'n':
                  lmt_ntwrk = strtoll(optarg, NULL, 0);
                  break;
#endif
            case 'p':
                  lmt_prcs = strtoll(optarg, NULL, 0);
                  break;
            case 'r':
                  lmt_rss = strtoll(optarg, NULL, 0);
                  break;
            case 's':
                  lmt_stck = strtoll(optarg, NULL, 0);
                  break;
            case 't':
                  lmt_tm = strtoll(optarg, NULL, 0);
                  break;
            case 'v':
                  vrbs = true;
                  break;
            case 'z':
                  lmt_fl_sz = strtoll(optarg, NULL, 0);
                  break;
            case 'h':
                  // FALLTHRU
            default:
                  cerr << HELP_MESSAGE;
                  return 1;
            }
      }
      argc -= optind;
      argv += optind;

      if (NULL == argv || NULL == *argv) {
            cerr << HELP_MESSAGE;
            return 1;
      }

      if (vrbs)
            cerr << "Resource limits:" << endl
                   << "      RLIMIT_AS           0x" << hex << lmt_all << " bytes" << endl
                   << "      RLIMIT_CORE         0x" << hex << lmt_cr << " bytes" << endl
                   << "      RLIMIT_DATA         0x" << hex << lmt_dta << " bytes" << endl
                   << "      RLIMIT_NOFILE       " << dec << lmt_fls << " files" << endl
                   << "      RLIMIT_FSIZE        0x" << hex << lmt_fl_sz << " bytes" << endl
                   << "      RLIMIT_MEMLOCK      0x" << hex << lmt_mlck << " bytes" << endl
#ifndef linux
                   << "      RLIMIT_SBSIZE       0x" << hex << lmt_ntwrk << " bytes" << endl
#endif
                   << "      RLIMIT_NPROC        " << dec << lmt_prcs << " processes" << endl
                   << "      RLIMIT_RSS          0x" << hex << lmt_rss << " bytes" << endl
                   << "      RLIMIT_STACK        0x" << hex << lmt_stck << " bytes" << endl
                   << "      RLIMIT_CPU          " << dec << lmt_tm << " seconds" << endl
                   ;


      /*
       * Create a random UID for the isolator.
       */
      isolator = random_uid();
      invoker = getuid();


      /*
       * Create the confinement directory. It should be safe to chown after
       * creating, because create_confinement_root fails unless 0:0 owns
       * CONFINEMENT_ROOT and it is the right mode.
       *
       * Alternatively, we could create confinement_path as the isolator,
       * but then we'd have to make CONFINEMENT_ROOT world-writable. I like
       * that less.
       */
      create_confinement_root();
      confinement_path = CONFINEMENT_ROOT + '/' + to_string(isolator, 16);

      if (mkdir(confinement_path.c_str(), 0755)) {
            throw runtime_error("Could not create " + confinement_path + ": " + strerror(errno));
      }

      if (lchown(confinement_path.c_str(), isolator, isolator)) {
            throw runtime_error("Could not lchown " + confinement_path + ": " + strerror(errno));
      }


      /*
       * Fork to do the isolation.
       */
      pid_t chld = fork();
      if (-1 == chld) {
            throw runtime_error("Could not fork");
      }

      int sts = 0;
      if (0 == chld) {

            drop_privilege_temporarily(isolator);

            string fl_pth = which(*argv);
            string pth = fl_pth;


            /*
             * Find out if the requested program is a script, and count the
             * dependencies.
             */

            FILE *fl = fopen(fl_pth.c_str(), "rb");
            if (shell_magic(fl)) {
                  char * pth = NULL;
                  size_t cpcty = 0;
                  ssize_t r = getline(&pth, &cpcty, fl);
                  (void) r;

                  /*
                   * Trim whitespace at the beginning and end of the
                   * interpreter path.
                   */
                  char *p = pth;
                  while (' ' == *p) {
                        p++;
                  }
                  char *q = strchr(p, '\n');
                  if (NULL == q) {
                        q = strchr(p, ' ');
                  }
                  *q = '\0';

                  lmt_fls += count_dependencies(p) + 5;

                  free(pth);

            } else {
                  /*
                   * Figure out what the requested program links to, and
                   * count those dependencies.
                   */
                  lmt_fls += count_dependencies(fl_pth);
            }
            fclose(fl);

            restore_privilege();


            /*
             * Do the final sandboxing.
             */
            restore_privilege();
            ensure_no_processes();
            if (chdir(confinement_path.c_str())) {
                  throw runtime_error("Could not confine to " + confinement_path + ": "
                        + strerror(errno));
            }


            /*
             * Set the resource limits.
             */
            drop_privilege_temporarily(invoker);
            set_resource_limit(RLIMIT_AS, lmt_all);
            set_resource_limit(RLIMIT_CORE, lmt_cr);
            set_resource_limit(RLIMIT_DATA, lmt_dta);
            set_resource_limit(RLIMIT_NOFILE, lmt_fls);
            set_resource_limit(RLIMIT_FSIZE, lmt_fl_sz);
            set_resource_limit(RLIMIT_MEMLOCK, lmt_mlck);
#ifndef linux
            set_resource_limit(RLIMIT_SBSIZE, lmt_ntwrk);
#endif
#ifdef linux
            /*
             * On Linux, setresuid(2) can fail (EAGAIN) because there
             * weren't enough processes. Kinky!
             */
            set_resource_limit(RLIMIT_NPROC, lmt_prcs + 1);
#else
            set_resource_limit(RLIMIT_NPROC, lmt_prcs);
#endif
            set_resource_limit(RLIMIT_RSS, lmt_rss);
            set_resource_limit(RLIMIT_STACK, lmt_stck);
            set_resource_limit(RLIMIT_CPU, lmt_tm);
            restore_privilege();


            /*
             * Whack the supplemental groups.
             */
            if (setgroups(0, NULL))
                  cerr << "Could not whack supplemental groups: " << strerror(errno) << endl;


            /*
             * Become the isolator, irrevocably.
             */
            drop_privilege_permanently(isolator);


            /*
             * Set the resource limits to their new minima. Previously,
             * they were set to root's minima (at the top of main), but we
             * do this again here with the new, low-privilege values.
             */
            lmt_all = min(lmt_all, get_resource_limit(RLIMIT_AS));
            lmt_cr = min(lmt_cr, get_resource_limit(RLIMIT_CORE));
            lmt_dta = min(lmt_dta, get_resource_limit(RLIMIT_DATA));
            lmt_fls = min(lmt_fls, get_resource_limit(RLIMIT_NOFILE));
            lmt_fl_sz = min(lmt_fl_sz, get_resource_limit(RLIMIT_FSIZE));
            lmt_mlck = min(lmt_mlck, get_resource_limit(RLIMIT_MEMLOCK));
#ifndef linux
            lmt_ntwrk = min(lmt_ntwrk, get_resource_limit(RLIMIT_SBSIZE));
#endif
            lmt_prcs = min(lmt_prcs, get_resource_limit(RLIMIT_NPROC));
            lmt_rss = min(lmt_rss, get_resource_limit(RLIMIT_RSS));
            lmt_stck = min(lmt_stck, get_resource_limit(RLIMIT_STACK));
            lmt_tm = min(lmt_tm, get_resource_limit(RLIMIT_CPU));


            /*
             * We are ready for launch.
             */

            char * const nvrnmnt [] = {
                  /*
                   * We need to set this so that the loader can find libs in
                   * /usr/local/lib.
                   */
                  (char *) "LD_LIBRARY_PATH=/lib:/usr/lib:/usr/local/lib",

                  (char *) ("DISPLAY=" + dsply).c_str(),
                  (char *) ("XAUTHORITY=" + xauthrty).c_str(),
                  (char *) ("HOME=" + confinement_path).c_str(),
                  (char *) "LOGNAME=isolator",
                  (char *) "USER=isolator",
                  (char *) "TERM=vt100",
                  NULL
            };

            execve(fl_pth.c_str(), argv, nvrnmnt);
            cerr << "Could not execute " << *argv << ": " << strerror(errno) << endl;
            _exit(errno);
      }
      else {
            (void) wait(&sts);

            if (vrbs) {
                  cerr << endl << "Resource usage statistics for \"";
                  for (int i = 0; i < argc; i++) {
                        cerr << argv[i] << (argc - 1 == i ? "" : " ");
                  }
                  cerr << "\": " << endl;
                  print_rusage();
            }

            if (unlink_recursively(confinement_path.c_str(), open(CONFINEMENT_ROOT.c_str(), O_RDONLY)))
                  cerr << confinement_path << ": " << strerror(errno) << endl;
      }

      return WEXITSTATUS(sts);
}

